from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import AllowedMetaData, Corpus, EntityType, MetaData, MetaType
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestMetaData(FixtureAPITestCase):
    """
    Tests related to Metadata and creation limitations from API
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.admin = User.objects.get(is_admin=True)
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.private_vol = cls.private_corpus.elements.create(type=cls.vol.type, name='Vol')
        cls.allowed_meta = AllowedMetaData.objects.bulk_create(
            AllowedMetaData(corpus=cls.corpus, type=t, name=n)
            for t, n in (
                (MetaType.Date, 'edition'),
                (MetaType.Text, 'edition'),
                (MetaType.Reference, '_id'),
            )
        )
        cls.worker_version = WorkerVersion.objects.first()
        cls.worker_run = cls.worker_version.worker_runs.first()

    def setUp(self):
        super().setUp()
        self.metadata = self.vol.metadatas.create(type=MetaType.Text, name='folio', value='123')
        self.private_metadata = self.private_vol.metadatas.create(type=MetaType.Markdown, name='leet', value='1337')

    def test_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'id': str(self.metadata.id),
                'name': 'folio',
                'type': 'text',
                'value': '123',
                'dates': [],
                'entity': None,
                'worker_version': None,
                'worker_run_id': None
            },
        ])

    def test_list_public(self):
        self.assertTrue(self.corpus.public)
        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'id': str(self.metadata.id),
                'name': 'folio',
                'type': 'text',
                'value': '123',
                'dates': [],
                'entity': None,
                'worker_version': None,
                'worker_run_id': None,
            },
        ])

    def test_list_private_requires_login(self):
        response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.private_vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_with_entity(self):
        # Link an entity to metadata, db requests count should remains same
        entity = self.corpus.entities.create(
            type=EntityType.Person,
            name='Marc',
            worker_version=self.worker_version,
        )
        entity_meta = self.vol.metadatas.create(
            name='meta',
            type=MetaType.Text,
            value='text_metadata',
            entity=entity,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json(), [
            {
                'dates': [],
                'id': str(self.metadata.id),
                'name': 'folio',
                'type': 'text',
                'value': '123',
                'entity': None,
                'worker_version': None,
                'worker_run_id': None
            },
            {
                'id': str(entity_meta.id),
                'name': 'meta',
                'type': 'text',
                'value': 'text_metadata',
                'dates': [],
                'entity': {
                    'dates': [],
                    'id': str(entity.id),
                    'metas': None,
                    'name': 'Marc',
                    'type': 'person',
                    'validated': False,
                    'worker_version_id': str(self.worker_version.id),
                    'worker_run': None,
                },
                'worker_version': None,
                'worker_run_id': None,
            },
        ])

    def test_list_with_worker_version(self):
        self.metadata.worker_version = self.worker_version
        self.metadata.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'dates': [],
                'id': str(self.metadata.id),
                'name': 'folio',
                'type': 'text',
                'value': '123',
                'entity': None,
                'worker_version': str(self.worker_version.id),
                'worker_run_id': None
            },
        ])

    def test_list_wrong_acl(self):
        self.vol.corpus = self.private_corpus
        self.vol.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_wrong_element_id(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:element-metadata', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_without_parents(self):
        self.vol.metadatas.create(type=MetaType.Numeric, name='weigth', value='1337')
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:element-metadata', kwargs={'pk': str(self.page.id)}),
                {'load_parents': False},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [{
            'dates': [],
            'entity': None,
            'id': str(self.page.metadatas.get().id),
            'name': 'folio',
            'type': 'text',
            'value': '1r',
            'worker_run_id': None,
            'worker_version': None,
        }])

    def test_list_with_parents(self):
        """
        It is possible to aggregate parents metadata by setting the `load_parents` flag. Requested element metadata
        will always been displayed first. In this case, page's folio is returned and volume's folio is evicted.
        """
        weight_meta = self.vol.metadatas.create(type=MetaType.Numeric, name='weigth', value='1337.0')
        self.client.force_login(self.user)
        # One more request is required to list element paths
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:element-metadata', kwargs={'pk': str(self.page.id)}),
                {'load_parents': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        def metadata_ordering(pk):
            md = MetaData.objects.get(pk=pk)
            return (md.element_id, md.name, md.id)
        self.assertListEqual(
            response.json(),
            sorted(
                [
                    {
                        'id': str(self.metadata.id),
                        'name': 'folio',
                        'type': 'text',
                        'value': '123',
                        'dates': [],
                        'entity': None,
                        'worker_version': None,
                        'worker_run_id': None,
                    },
                    {
                        'dates': [],
                        'id': str(self.page.metadatas.get().id),
                        'name': 'folio',
                        'type': 'text',
                        'value': '1r',
                        'entity': None,
                        'worker_version': None,
                        'worker_run_id': None,
                    },
                    {
                        'dates': [],
                        'id': str(weight_meta.id),
                        'entity': None,
                        'name': 'weigth',
                        'type': 'numeric',
                        'value': 1337.0,
                        'worker_run_id': None,
                        'worker_version': None,
                    },
                ],
                key=lambda x: metadata_ordering(x["id"]),
            )
        )

    def test_metadata_edit_no_post(self):
        """
        Only POST method is forbidden
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_create_metadata_verified(self):
        response = self.client.post(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_metadata_methods(self):
        self.client.force_login(self.user)
        methods = (self.client.patch, self.client.put)
        for method in methods:
            response = method(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_create_metadata_writable_element(self):
        self.client.force_login(self.user)
        hidden_book = self.private_corpus.elements.create(
            type=self.corpus.types.get(slug='volume'),
            name='Nope'
        )
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(hidden_book.id)}),
            data={'type': 'text', 'name': 'language', 'value': 'Japanese'}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_metadata(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Location, name='location')
        self.assertEqual(md.value, 'Texas')

    def test_create_metadata_worker_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas', 'worker_version': str(self.worker_version.id)}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Location, name='location')
        self.assertEqual(md.value, 'Texas')
        self.assertEqual(md.worker_version, self.worker_version)

    def test_create_metadata_worker_run(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas', 'worker_run_id': str(self.worker_run.id)}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Location, name='location')
        self.assertEqual(md.value, 'Texas')
        self.assertEqual(md.worker_run_id, self.worker_run.id)
        self.assertEqual(md.worker_version, self.worker_version)

    def test_create_metadata_worker_run_or_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas', 'worker_run_id': str(self.worker_run.id), 'worker_version': str(self.worker_version.id)}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'non_field_errors': ['Only one of `worker_version` and `worker_run_id` may be set.']})

    def test_create_metadata_bad_worker_run(self):
        self.client.force_login(self.user)
        non_existent_worker_run_id = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas', 'worker_run_id': non_existent_worker_run_id},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'worker_run_id': [f'Invalid pk "{non_existent_worker_run_id}" - object does not exist.']})

    def test_create_metadata_date(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'date', 'name': 'date', 'value': '1885'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        metadata_id = self.vol.metadatas.get(type=MetaType.Date, name='date').id
        self.assertDictEqual(response.json(), {
            'id': str(metadata_id),
            'name': 'date',
            'type': 'date',
            'value': '1885',
            'entity': None,
            'dates': [{
                'day': None,
                'month': None,
                'type': 'exact',
                'year': 1885
            }],
            'worker_version': None,
            'worker_run_id': None
        })

    def test_create_metadata_empty(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'name': ['This field is required.'],
            'type': ['This field is required.'],
            'value': ['This field is required.']
        })

    def test_create_metadata_index(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'date', 'name': 'edition', 'value': '1986-june'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'text', 'name': 'edition', 'value': 'Hammersmith youth edition'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        metadata = self.vol.metadatas.filter(name='edition')
        self.assertEqual(metadata.count(), 2)

    def test_create_duplicated_metadata(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
                data={'type': 'date', 'name': 'edition', 'value': '1986-june'}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        metadata_id = response.json()['id']

        with self.assertNumQueries(6):
            response = self.client.post(
                reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
                data={'type': 'date', 'name': 'edition', 'value': '1986-june'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            'detail': ['This metadata already exists.'],
            'id': metadata_id,
        })

        metadata = self.vol.metadatas.filter(name='edition')
        self.assertEqual(metadata.count(), 1)

    def test_create_duplicated_metadata_exception(self):
        """
        The DB cannot have a unique constraint in place for element/name/type/value, and the APIs are supposed to deal
        with that by checking themselves before inserting. If we still find duplicate metadata in the database, the API
        should fail to allow us to notice the issue and troubleshoot it.
        """
        self.vol.metadatas.create(name='test', type=MetaType.Text, value='And again')
        self.vol.metadatas.create(name='test', type=MetaType.Text, value='And again')
        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            with self.assertRaises(MetaData.MultipleObjectsReturned):
                self.client.post(
                    reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
                    data={'name': 'test', 'type': 'text', 'value': 'And again'},
                )

    def test_delete_metadata_verified(self):
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_metadata(self):
        self.client.force_login(self.user)
        self.assertEqual(self.vol.metadatas.count(), 1)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(self.vol.metadatas.count(), 0)

    def test_delete_non_allowed_metadata(self):
        """
        A user may not be able to delete a non allowed metadata
        """
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Edition of metadata with this type (Location) and name (not editable) is not allowed for this corpus']}
        )

    def test_admin_delete_any(self):
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.admin)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_metadata_private_corpus(self):
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.private_metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_delete_metadata_readable_element(self):
        """
        An explicit message should be raised when user can read but not delete metadata
        """
        self.client.force_login(self.user)
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        response = self.client.delete(reverse('api:metadata-edit', kwargs={'pk': str(self.private_metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have write access to this corpus.'})

    def test_admin_create_any(self):
        """
        Admin and internal users are allowed to create any metadata
        """
        for user in (self.admin, self.internal_user):
            with self.subTest(user=user):
                self.client.force_login(user)
                response = self.client.post(
                    reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
                    data={'type': 'text', 'name': 'certainly not allowed', 'value': 'bla bla bla'}
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED)
                md = self.vol.metadatas.get(type=MetaType.Text, name='certainly not allowed')
                self.assertEqual(md.value, 'bla bla bla')
                md.delete()
                allowed_md = self.vol.corpus.allowed_metadatas.filter(name='certainly not allowed', type=MetaType.Text)
                self.assertTrue(allowed_md.exists())

    def test_create_non_allowed_metadata(self):
        """
        Non internal user is limited in metadata creation
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'text', 'name': 'certainly not allowed', 'value': 'bla bla bla'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Metadata with this type (Text) and name (certainly not allowed) are not allowed for this corpus']}
        )

    def test_update_metadata(self):
        self.client.force_login(self.user)
        self.assertEqual(self.metadata.type, MetaType.Text)
        self.assertEqual(self.metadata.value, "123")
        self.assertEqual(self.metadata.name, "folio")
        response = self.client.put(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date',
                'name': 'date',
                'value': '2019-12-04'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'dates': [{'day': 4, 'month': 12, 'type': 'exact', 'year': 2019}],
            'id': str(self.metadata.id),
            'name': 'date',
            'type': 'date',
            'value': '2019-12-04',
            'entity': None,
            'worker_version': None,
            'worker_run_id': None
        })
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.type, MetaType.Date)
        self.assertEqual(self.metadata.name, 'date')
        self.assertEqual(self.metadata.value, '2019-12-04')

    def test_update_metadata_incomplete(self):
        """
        UpdateMetaData requires type, name and value to be set
        """
        self.client.force_login(self.user)
        self.assertEqual(self.metadata.type, MetaType.Text)
        response = self.client.put(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date',
                'name': 'date',
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"value": ["This field is required."]})

    def test_update_metadata_disallowed(self):
        self.client.force_login(self.user)
        response = self.client.put(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date',
                'name': 'death of William I, Count of Hainaut',
                'value': '1337'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Edition of metadata with this type (Date) and name (death of William I, Count of Hainaut) is not allowed for this corpus']}
        )

    def test_update_metadata_from_any_internal(self):
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.internal_user)
        response = self.client.put(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': 'I got you', 'value': 'patched !'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        metadata.refresh_from_db()
        self.assertEqual(metadata.value, 'patched !')

    def test_update_metadata_to_any_internal(self):
        self.client.force_login(self.internal_user)
        response = self.client.put(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={'type': 'text', 'name': 'death of William I, Count of Hainaut', 'value': '1337'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.name, 'death of William I, Count of Hainaut')
        self.assertEqual(self.metadata.value, '1337')

    def test_update_metadata_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.put(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'joke',
                'name': 'what is my name ?',
                'value': 'i have no name, im a machine'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'type': ['Value is not of type MetaType']})

    def test_update_metadata_duplicated(self):
        """
        UpdateMetaData should not allow creating duplicated metadata
        """
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='location', value='111,222')
        self.client.force_login(self.user)
        response = self.client.put(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': self.metadata.name, 'value': self.metadata.value}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'detail': ['This metadata already exists.'],
            'id': str(self.metadata.id),
        })

    def test_update_metadata_worker_version(self):
        """
        UpdateMetaData should not allow updating the worker version ID
        """
        self.assertIsNone(self.metadata.worker_version_id)
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
                data={'name': 'hello', 'type': 'text', 'value': 'buenos dias', 'worker_version': str(self.worker_version.id)}
            )
            # DRF should ignore the field silently
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertIsNone(data['worker_version'])
        self.metadata.refresh_from_db()
        self.assertIsNone(self.metadata.worker_version_id)
        self.assertEqual(self.metadata.name, 'hello')
        self.assertEqual(self.metadata.type, MetaType.Text)
        self.assertEqual(self.metadata.value, 'buenos dias')

    def test_patch_metadata(self):
        self.client.force_login(self.user)
        self.assertEqual(self.metadata.type, MetaType.Text)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date',
                'name': 'date',
                'value': '2019-12-04'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'dates': [{'day': 4, 'month': 12, 'type': 'exact', 'year': 2019}],
            'id': str(self.metadata.id),
            'name': 'date',
            'type': 'date',
            'value': '2019-12-04',
            'entity': None,
            'worker_version': None,
            'worker_run_id': None
        })
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.type, MetaType.Date)
        self.assertEqual(self.metadata.name, 'date')
        self.assertEqual(self.metadata.value, '2019-12-04')

    def test_patch_metadata_type(self):
        self.client.force_login(self.user)
        self.assertEqual(self.metadata.type, MetaType.Text)
        self.corpus.allowed_metadatas.create(type=MetaType.Date, name='folio')
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())
        self.assertDictEqual(response.json(), {
            'dates': [],
            'id': str(self.metadata.id),
            'name': 'folio',
            'type': 'date',
            'value': '123',
            'entity': None,
            'worker_version': None,
            'worker_run_id': None
        })
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.type, MetaType.Date)
        self.assertEqual(self.metadata.name, 'folio')
        self.assertEqual(self.metadata.value, '123')

    def test_patch_metadata_value(self):
        self.client.force_login(self.user)
        self.assertEqual(self.metadata.type, MetaType.Text)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'value': '2019-12-04'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())
        self.assertDictEqual(response.json(), {
            'dates': [],
            'id': str(self.metadata.id),
            'name': 'folio',
            'type': 'text',
            'value': '2019-12-04',
            'entity': None,
            'worker_version': None,
            'worker_run_id': None
        })
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.type, MetaType.Text)
        self.assertEqual(self.metadata.name, 'folio')
        self.assertEqual(self.metadata.value, '2019-12-04')

    def test_patch_disallowed(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'date',
                'name': 'death of William I, Count of Hainaut',
                'value': '1337'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Edition of metadata with this type (Date) and name (death of William I, Count of Hainaut) is not allowed for this corpus']}
        )

    def test_patch_partial_disallowed(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'name': 'death of William I, Count of Hainaut',
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Edition of metadata with this type (Text) and name (death of William I, Count of Hainaut) is not allowed for this corpus']}
        )

    def test_patch_from_any_internal(self):
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': 'I got you', 'value': 'patched !'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        metadata.refresh_from_db()
        self.assertEqual(metadata.value, 'patched !')

    def test_patch_to_any_internal(self):
        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={'type': 'text', 'name': 'death of William I, Count of Hainaut', 'value': '1337'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.metadata.refresh_from_db()
        self.assertEqual(self.metadata.name, 'death of William I, Count of Hainaut')
        self.assertEqual(self.metadata.value, '1337')

    def test_patch_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
            data={
                'type': 'joke',
                'name': 'what is my name ?',
                'value': 'i have no name, im a machine'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'type': ['Value is not of type MetaType']})

    def test_patch_to_existing_metadata(self):
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='location', value='111,222')
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': 'folio', 'value': '333,444'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        metadata.refresh_from_db()
        self.assertEqual(metadata.value, '333,444')

    def test_patch_unallowed_metadata(self):
        """
        A user may not be able to patch a non allowed metadata
        """
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='not editable', value='111,222')
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'type': 'text', 'name': 'folio', 'value': '333,444'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            '__all__': ['Edition of metadata with this type (Location) and name (not editable) is not allowed for this corpus']
        })
        metadata.refresh_from_db()
        self.assertEqual(metadata.name, 'not editable')
        self.assertEqual(metadata.value, '111,222')

    def test_patch_worker_version(self):
        """
        PartialUpdateMetaData should not allow updating the worker version ID
        """
        self.assertIsNone(self.metadata.worker_version_id)
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}),
                data={'worker_version': str(self.worker_version.id)}
            )
            # DRF should ignore the field silently
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertIsNone(data['worker_version'])
        self.metadata.refresh_from_db()
        self.assertIsNone(self.metadata.worker_version_id)

    def test_patch_duplicated(self):
        """
        PartialUpdateMetaData should not allow creating duplicated metadata
        """
        metadata = self.vol.metadatas.create(type=MetaType.Text, name=self.metadata.name, value='some value')
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'value': self.metadata.value}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'detail': ['This metadata already exists.'],
            'id': str(self.metadata.id),
        })

    def test_get_metadata(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                'id': str(self.metadata.id),
                'type': 'text',
                'name': 'folio',
                'value': '123',
                'entity': None,
                'dates': [],
                'worker_version': None,
                'worker_run_id': None
            }
        )

    def test_get_metadata_numeric(self):
        """
        Numeric metadata should have values serialized as floats
        """
        self.metadata.type = MetaType.Numeric
        self.metadata.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(self.metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                'id': str(self.metadata.id),
                'type': 'numeric',
                'name': 'folio',
                'value': 123.0,
                'entity': None,
                'dates': [],
                'worker_version': None,
                'worker_run_id': None
            }
        )

    def test_create_metadata_entity(self):
        self.client.force_login(self.superuser)
        entity = self.corpus.entities.create(
            name='Texas',
            type=EntityType.Location,
            worker_version=self.worker_version,
        )
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas', 'entity_id': entity.id}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        self.assertEqual(self.vol.metadatas.get(name='location').entity, entity)

    def test_patch_metadata_entity(self):
        self.client.force_login(self.superuser)
        entity = self.corpus.entities.create(
            name='Texas',
            type=EntityType.Location,
            worker_version=self.worker_version,
        )
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='location', value='Texas')
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'entity_id': entity.id}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())
        metadata.refresh_from_db()
        self.assertEqual(metadata.entity, entity)

    def test_patch_metadata_entity_none(self):
        self.client.force_login(self.superuser)
        entity = self.corpus.entities.create(
            name='Texas',
            type=EntityType.Location,
            worker_version=self.worker_version,
        )
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='location', value='Texas', entity=entity)
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'entity_id': None},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())
        metadata.refresh_from_db()
        self.assertIsNone(metadata.entity)

    def test_create_metadata_entity_corpus_check(self):
        self.client.force_login(self.superuser)
        entity = self.private_corpus.entities.create(
            name='Texas',
            type=EntityType.Location,
            worker_version=self.worker_version,
        )
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas', 'entity_id': entity.id}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_patch_metadata_entity_corpus_check(self):
        self.client.force_login(self.superuser)
        entity = self.private_corpus.entities.create(
            name='Texas',
            type=EntityType.Location,
            worker_version=self.worker_version,
        )
        metadata = self.vol.metadatas.create(type=MetaType.Location, name='location', value='Texas')
        response = self.client.patch(
            reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}),
            data={'entity_id': entity.id}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        metadata.refresh_from_db()
        self.assertIsNone(metadata.entity)

    def test_render_markdown_metadata(self):
        metadata = self.vol.metadatas.create(type=MetaType.Markdown, name='Some text', value='# Title\n## Subtitle\nbla')
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                'id': str(metadata.id),
                'type': 'markdown',
                'name': 'Some text',
                'value': '# Title\n## Subtitle\nbla',
                'entity': None,
                'dates': [],
                'worker_version': None,
                'worker_run_id': None
            }
        )

    def test_render_html_metadata(self):
        metadata = self.vol.metadatas.create(type=MetaType.Markdown, name='Some text', value='<h1>Title</h1>')
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                'id': str(metadata.id),
                'type': 'markdown',
                'name': 'Some text',
                'value': '<h1>Title</h1>',
                'entity': None,
                'dates': [],
                'worker_version': None,
                'worker_run_id': None
            }
        )

    def test_unrender_html_metadata(self):
        metadata = self.vol.metadatas.create(type=MetaType.Markdown, name='Some text', value='<style type="text/css">* { display: none !important; }</style>')
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(metadata.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            response.json(),
            {
                'id': str(metadata.id),
                'type': 'markdown',
                'name': 'Some text',
                'value': '&lt;style type="text/css"&gt;* { display: none !important; }&lt;/style&gt;',
                'entity': None,
                'dates': [],
                'worker_version': None,
                'worker_run_id': None
            }
        )

    def test_create_markdown_metadata(self):
        AllowedMetaData.objects.create(corpus=self.corpus, type=MetaType.Markdown, name='text')
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'markdown', 'name': 'text', 'value': '# Title\n## Subtitle\nbla'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Markdown, name='text')
        self.assertEqual(md.value, '# Title\n## Subtitle\nbla')
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(md.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['value'], '# Title\n## Subtitle\nbla')

    def test_create_html_metadata(self):
        AllowedMetaData.objects.create(corpus=self.corpus, type=MetaType.Markdown, name='text')
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'markdown', 'name': 'text', 'value': '<style type="text/css">* { display: none !important; }</style>'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Markdown, name='text')
        self.assertEqual(md.value, '<style type="text/css">* { display: none !important; }</style>')
        response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(md.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['value'], '&lt;style type="text/css"&gt;* { display: none !important; }&lt;/style&gt;')

    def test_create_html_metadata_with_other_type(self):
        """
        Creating a metadata of a type other than Markdown should not cause any HTML sanitization
        """
        self.client.force_login(self.superuser)
        for type in {MetaType.Text, MetaType.Date, MetaType.Location, MetaType.Reference}:
            with self.subTest(type=type):
                response = self.client.post(
                    reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
                    data={'type': type.value, 'name': 'edition', 'value': '<a href="lalaland">oops</a>'}
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())

                md = self.vol.metadatas.get(type=type, name='edition')
                self.assertEqual(md.value, '<a href="lalaland">oops</a>')

                response = self.client.get(reverse('api:metadata-edit', kwargs={'pk': str(md.id)}))
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertEqual(response.json()['value'], '<a href="lalaland">oops</a>')

    def test_create_numeric_metadata(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'numeric', 'name': 'blah', 'value': '13.37'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Numeric, name='blah')
        self.assertEqual(md.value, '13.37')

    def test_create_numeric_metadata_invalid(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'numeric', 'name': 'blah', 'value': 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'value': ['The value of a numeric metadata should be a valid floating-point number.']})

    def test_create_url_metadata(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'url', 'name': 'blah', 'value': 'http://nowhere.com'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.URL, name='blah')
        self.assertEqual(md.value, 'http://nowhere.com')

    def test_create_url_metadata_invalid(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'url', 'name': 'blah', 'value': 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'value': ['The value of a URL metadata must be a valid HTTP or HTTPS URL.']})

    def test_create_url_metadata_scheme(self):
        self.client.force_login(self.superuser)
        response = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'url', 'name': 'blah', 'value': 'git+https://gitlab.com/teklia/arkindex/backend.git'}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'value': ['The value of a URL metadata must be a valid HTTP or HTTPS URL.']})

    def test_bulk_create_metadata_verified(self):
        response = self.client.post(reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_create_metadata_methods(self):
        self.client.force_login(self.user)
        methods = (self.client.patch, self.client.put, self.client.get, self.client.delete)
        for method in methods:
            response = method(reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_bulk_create_metadata_writable_element(self):
        self.client.force_login(self.user)
        hidden_book = self.private_corpus.elements.create(
            type=self.corpus.types.get(slug='volume'),
            name='Nope'
        )
        response = self.client.post(
            reverse('api:element-metadata-bulk', kwargs={'pk': str(hidden_book.id)}),
            data={'metadata_list': [
                {'type': 'text', 'name': 'language', 'value': 'Japanese'}
            ]},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_create_metadata(self):
        entity = self.corpus.entities.create(name='42', type=EntityType.Number)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
            data={'metadata_list': [
                {'type': 'location', 'name': 'location', 'value': 'Texas'},
                {'type': 'date', 'name': 'date', 'value': '1885'},
                {'type': 'numeric', 'name': 'numeric', 'value': '42', 'entity_id': str(entity.id)}
            ]},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.vol.metadatas.count(), 4)  # 1 added during setUp
        md1 = self.vol.metadatas.get(type=MetaType.Location, name='location')
        self.assertEqual(md1.value, 'Texas')
        md2 = self.vol.metadatas.get(type=MetaType.Date, name='date')
        self.assertEqual(md2.value, '1885')
        md3 = self.vol.metadatas.get(type=MetaType.Numeric, name='numeric')
        self.assertEqual(md3.value, '42')
        self.assertEqual(md3.entity, entity)

        self.assertDictEqual(response.json(), {
            'metadata_list': [
                {
                    'id': str(md1.id),
                    'name': 'location',
                    'type': 'location',
                    'value': 'Texas',
                    'entity_id': None,
                    'dates': []
                }, {
                    'id': str(md2.id),
                    'name': 'date',
                    'type': 'date',
                    'value': '1885',
                    'entity_id': None,
                    'dates': [{
                        'day': None,
                        'month': None,
                        'type': 'exact',
                        'year': 1885
                    }]
                }, {
                    'id': str(md3.id),
                    'name': 'numeric',
                    'type': 'numeric',
                    'value': 42.0,
                    'entity_id': str(entity.id),
                    'dates': []
                }
            ],
            'worker_version': None,
            'worker_run_id': None
        })

    def test_bulk_create_metadata_worker_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
            data={'metadata_list': [
                {'type': 'text', 'name': 'language', 'value': 'Chinese'},
                {'type': 'text', 'name': 'alt_language', 'value': 'Mandarin'}
            ], 'worker_version': str(self.worker_version.id)},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Text, name='language')
        self.assertEqual(md.value, 'Chinese')
        self.assertEqual(md.worker_version, self.worker_version)
        md = self.vol.metadatas.get(type=MetaType.Text, name='alt_language')
        self.assertEqual(md.value, 'Mandarin')
        self.assertEqual(md.worker_version, self.worker_version)

    def test_bulk_create_metadata_worker_run(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
            data={'metadata_list': [
                {'type': 'text', 'name': 'language', 'value': 'Chinese'},
                {'type': 'text', 'name': 'alt_language', 'value': 'Mandarin'}
            ], 'worker_run_id': str(self.worker_run.id)},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        md = self.vol.metadatas.get(type=MetaType.Text, name='language')
        self.assertEqual(md.value, 'Chinese')
        self.assertEqual(md.worker_run_id, self.worker_run.id)
        self.assertEqual(md.worker_version, self.worker_version)
        md = self.vol.metadatas.get(type=MetaType.Text, name='alt_language')
        self.assertEqual(md.value, 'Mandarin')
        self.assertEqual(md.worker_run_id, self.worker_run.id)
        self.assertEqual(md.worker_version, self.worker_version)

    def test_bulk_create_metadata_worker_run_or_version(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
            data={'metadata_list': [
                {'type': 'text', 'name': 'language', 'value': 'Chinese'},
                {'type': 'text', 'name': 'alt_language', 'value': 'Mandarin'}
            ], 'worker_run_id': str(self.worker_run.id), 'worker_version': str(self.worker_version.id)},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'non_field_errors': ['Only one of `worker_version` and `worker_run_id` may be set.']})

    def test_bulk_create_metadata_bad_worker_run(self):
        self.client.force_login(self.user)
        non_existent_worker_run_id = "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
        response = self.client.post(
            reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
            data={'metadata_list': [
                {'type': 'text', 'name': 'language', 'value': 'Chinese'},
                {'type': 'text', 'name': 'alt_language', 'value': 'Mandarin'}
            ], 'worker_run_id': non_existent_worker_run_id},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'worker_run_id': [f'Invalid pk "{non_existent_worker_run_id}" - object does not exist.']})

    def test_bulk_create_metadata_empty(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_list': ['This field is required.']
        })

    def test_bulk_create_metadata_empty_list(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
            data={'metadata_list': []},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_list': {'non_field_errors': ['This list may not be empty.']}
        })

    def test_admin_bulk_create_any(self):
        """
        Admin and internal users are allowed to create any metadata
        """
        for user, setup in ((self.admin, True), (self.internal_user, False)):
            with self.subTest(user=user):
                self.client.force_login(user)
                response = self.client.post(
                    reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
                    data={'metadata_list': [
                        {'type': 'text', 'name': 'certainly not allowed', 'value': 'bla bla bla'},
                        {'type': 'location', 'name': 'not allowed', 'value': 'boo'},
                    ]},
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED)
                self.assertEqual(self.vol.metadatas.count(), 2 + setup)  # 1 added during setUp
                md1 = self.vol.metadatas.get(type=MetaType.Text, name='certainly not allowed')
                self.assertEqual(md1.value, 'bla bla bla')
                md2 = self.vol.metadatas.get(type=MetaType.Location, name='not allowed')
                self.assertEqual(md2.value, 'boo')

                self.assertDictEqual(response.json(), {
                    'metadata_list': [
                        {
                            'id': str(md1.id),
                            'name': 'certainly not allowed',
                            'type': 'text',
                            'value': 'bla bla bla',
                            'dates': [],
                            'entity_id': None,
                        }, {
                            'id': str(md2.id),
                            'name': 'not allowed',
                            'type': 'location',
                            'value': 'boo',
                            'dates': [],
                            'entity_id': None,
                        },
                    ],
                    'worker_version': None,
                    'worker_run_id': None
                })

                self.vol.metadatas.all().delete()

                allowed_md = self.vol.corpus.allowed_metadatas.filter(name__contains='not allowed')
                self.assertCountEqual(
                    [{'name': a.name, 'type': a.type} for a in allowed_md],
                    [
                        {'name': 'certainly not allowed', 'type': MetaType.Text},
                        {'name': 'not allowed', 'type': MetaType.Location}
                    ]
                )

    def test_bulk_create_metadata_unique(self):
        """
        The metadata created with the bulk endpoint must be unique
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_list': [
                    {'type': 'location', 'name': 'location', 'value': 'Texas'},
                    {'type': 'location', 'name': 'location', 'value': 'Texas'}
                ]},
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['Metadata in metadata list must make a unique set.']
        })

    def test_bulk_create_metadata_no_duplicates(self):
        """
        Do not create metadata that already exist on the element
        """
        self.client.force_login(self.user)
        response_1 = self.client.post(
            reverse('api:element-metadata', kwargs={'pk': str(self.vol.id)}),
            data={'type': 'location', 'name': 'location', 'value': 'Texas'}
        )
        self.assertEqual(response_1.status_code, status.HTTP_201_CREATED)
        with self.assertNumQueries(4):
            response_2 = self.client.post(
                reverse('api:element-metadata-bulk', kwargs={'pk': str(self.vol.id)}),
                data={'metadata_list': [
                    {'type': 'location', 'name': 'location', 'value': 'Texas'}
                ]},
                format='json'
            )
        self.assertEqual(response_2.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response_2.json(), {
            'non_field_errors': ["Metadata {('location', 'location', 'Texas')} already exist(s) on this element."]
        })
