from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.documents.models import Corpus, Element
from arkindex.process.models import Process, ProcessMode, WorkerConfiguration, WorkerVersion, WorkerVersionGPUUsage
from arkindex.project.mixins import TrainingModelMixin, WorkerACLMixin
from arkindex.training.models import Model, ModelVersion


class StartTrainingSerializer(serializers.ModelSerializer, WorkerACLMixin, TrainingModelMixin):

    name = serializers.CharField(max_length=100, write_only=True)
    corpus_id = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        help_text=(
            'ID of the corpus containing the test and train folders. '
            'The user must have a contributor access on the corpus.'
        ),
        style={'base_template': 'input.html'},
        write_only=True,
    )
    train_folder_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        help_text='ID of the folder element containing the train dataset.',
        style={'base_template': 'input.html'},
        write_only=True,
    )
    validation_folder_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        help_text='ID of the folder element containing the validation dataset.',
        style={'base_template': 'input.html'},
        required=False,
        allow_null=True,
        write_only=True,
    )
    test_folder_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        help_text='ID of the folder element containing the test dataset.',
        style={'base_template': 'input.html'},
        required=False,
        allow_null=True,
        write_only=True,
    )
    worker_version_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerVersion.objects.none(),
        help_text=(
            'ID of the worker version to publish a new trained model to. '
            'The user must have an admin access to the worker or repository.'
        ),
        style={'base_template': 'input.html'},
        write_only=True,
    )
    model_id = serializers.PrimaryKeyRelatedField(
        queryset=Model.objects.none(),
        help_text=(
            'ID of the model that will receive the new model version being created by the process. '
            'The user must have a contributor access to the model.'
        ),
        style={'base_template': 'input.html'},
        write_only=True,
    )
    # Use a UUIDField for the model version to ensure it matches the specified model
    model_version_id = serializers.UUIDField(
        help_text=(
            'Optional ID of a model version to be trained again. '
            'The model version must be part of the specified model.'
        ),
        style={'base_template': 'input.html'},
        required=False,
        allow_null=True,
        write_only=True,
    )
    worker_configuration_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerConfiguration.objects.all(),
        help_text='Optional worker configuration to apply on the worker being started',
        style={'base_template': 'input.html'},
        required=False,
        allow_null=True,
        write_only=True,
    )
    use_gpu = serializers.BooleanField(default=False, write_only=True)

    class Meta:
        model = Process
        fields = (
            'id',
            'name',
            'corpus_id',
            'train_folder_id',
            'validation_folder_id',
            'test_folder_id',
            'worker_version_id',
            'model_id',
            'model_version_id',
            'worker_configuration_id',
            'use_gpu',
        )
        read_only_fields = ('id',)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'request' not in self.context:
            return

        # This is a hack to make the ACLMixins work on this serializer,
        # because they are designed to be used on APIs and not serializers
        self._user = self.context['request'].user
        corpora = Corpus.objects.writable(self._user)

        # Ensure the user has a contributor access to the corpus and folders
        self.fields['corpus_id'].queryset = corpora
        elements_qs = Element.objects.filter(corpus__in=corpora, type__folder=True)
        self.fields['train_folder_id'].queryset = elements_qs
        self.fields['validation_folder_id'].queryset = elements_qs
        self.fields['test_folder_id'].queryset = elements_qs

        # Restrict the worker version field to all workers that the user can start
        self.fields['worker_version_id'].queryset = WorkerVersion.objects.filter(worker__in=self.executable_workers)

        # Ensure the user has a contributor access to a potential model
        self.fields['model_id'].queryset = self.editable_models

    def validate_worker_version_id(self, version):
        if not version.docker_image_id or not version.docker_image_iid:
            raise ValidationError(detail='This version has no Docker image, which is required for the training.')
        return version

    def validate(self, data):
        data = super().validate(data)

        # Ensure test, val and train folders are different
        train_check_id = data['train_folder_id'].id
        val_check_id = getattr(data.get('validation_folder_id'), 'id', 0)
        test_check_id = getattr(data.get('test_folder_id'), 'id', 1)
        if (
            val_check_id == train_check_id
            or test_check_id == train_check_id
            or val_check_id == test_check_id
        ):
            raise ValidationError({'__all__': ['Train, validation and test folders must be different.']})

        # Ensure the worker version is compatible with the GPU mode
        if data['worker_version_id'].gpu_usage == WorkerVersionGPUUsage.Disabled and data["use_gpu"]:
            raise ValidationError({'__all__': ['This worker version does not support GPU mode.']})
        elif data['worker_version_id'].gpu_usage == WorkerVersionGPUUsage.Required and not data["use_gpu"]:
            raise ValidationError({'__all__': ['This worker version does only support GPU mode.']})

        # Ensure the worker configuration match the given worker
        configuration = data.get('worker_configuration_id')
        if configuration and configuration.worker_id != data['worker_version_id'].worker_id:
            raise ValidationError({'__all__': ['The configuration is not compatible with this worker.']})

        # Ensure the model version is part of the defined model and exists
        if data.get("model_version_id"):
            try:
                data["model_version_id"] = data["model_id"].versions.get(id=data["model_version_id"])
            except ModelVersion.DoesNotExist:
                raise ValidationError({
                    'model_version_id': ['Object does not exist or is not a version of the specified model']
                })

        return data

    def create(self, validated_data):

        self.instance = super().create({
            'name': validated_data["name"],
            'corpus': validated_data["corpus_id"],
            'model': validated_data["model_id"],
            'train_folder': validated_data["train_folder_id"],
            'validation_folder': validated_data.get("validation_folder_id"),
            'test_folder': validated_data.get("test_folder_id"),
            'mode': ProcessMode.Training,
            'creator': self._user,
        })

        # Start the process with the extra args
        self.instance.start(
            worker_version=validated_data["worker_version_id"],
            model_version=validated_data.get("model_version_id"),
            worker_configuration=validated_data.get("worker_configuration_id"),
            use_gpu=validated_data["use_gpu"],
        )
        return self.instance
