import uuid
from unittest.mock import MagicMock, call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.process.models import (
    ActivityState,
    Process,
    ProcessMode,
    Repository,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.process.tasks import initialize_activity
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import User


class TestWorkerActivity(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.element = Element.objects.get(name='Volume 1, page 2r')
        cls.process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=cls.user
        )
        cls.configuration = WorkerConfiguration.objects.create(worker=cls.worker_version.worker, name='A config', configuration={'a': 'b'})
        cls.process.worker_runs.create(version=cls.worker_version, parents=[], configuration=cls.configuration)
        cls.worker_type = WorkerType.objects.get(slug='recognizer')

    def setUp(self):
        super().setUp()
        # Create a queued activity for this element
        self.activity = self.element.activities.create(
            process=self.process,
            worker_version=self.worker_version,
            state=WorkerActivityState.Queued,
            configuration=self.configuration
        )

    def test_bulk_insert_activity_no_configuration(self):
        """
        Bulk insert worker activities for acts
        """
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        params = {
            'worker_version_id': self.worker_version.id,
            'corpus_id': self.corpus.id,
            'state': WorkerActivityState.Started.value,
            'process_id': self.process.id,
        }
        with self.assertExactQueries('workeractivity_bulk_insert_no_configuration.sql', params=params):
            WorkerActivity.objects.bulk_insert(self.worker_version.id, self.process.id, None, elements_qs, state=WorkerActivityState.Started)
        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Started, process=self.process).count(), 5)

    def test_bulk_insert_activity_children(self):
        """
        Bulk insert worker activities for acts
        """
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        params = {
            'worker_version_id': self.worker_version.id,
            'configuration_id': self.configuration.id,
            'corpus_id': self.corpus.id,
            'state': WorkerActivityState.Started.value,
            'process_id': self.process.id,
        }
        with self.assertExactQueries('workeractivity_bulk_insert.sql', params=params):
            WorkerActivity.objects.bulk_insert(self.worker_version.id, self.process.id, self.configuration.id, elements_qs, state=WorkerActivityState.Started)
        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Started, process=self.process).count(), 5)

    def test_bulk_insert_activity_existing(self):
        """
        WorkerActivity.objects.bulk_insert should steal existing WorkerActivities from other processes
        if those are in a queued or error state
        """
        old_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user)
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        self.assertEqual(elements_qs.count(), 5)

        self.process.activities.all().delete()

        # Create one WorkerActivity for each possible state
        WorkerActivity.objects.bulk_create([
            WorkerActivity(
                element=element,
                worker_version=self.worker_version,
                state=state,
                process=old_process,
                configuration=self.configuration
            )
            for element, state in zip(elements_qs, WorkerActivityState)
        ])

        self.assertEqual(self.process.activities.count(), 0)
        self.assertEqual(old_process.activities.count(), 4)
        self.assertCountEqual(list(old_process.activities.values_list('state', flat=True).distinct()), [
            WorkerActivityState.Queued,
            WorkerActivityState.Started,
            WorkerActivityState.Processed,
            WorkerActivityState.Error,
        ])

        with self.assertNumQueries(1):
            WorkerActivity.objects.bulk_insert(
                self.worker_version.id,
                self.process.id,
                self.configuration.id,
                elements_qs,
                state=WorkerActivityState.Started,
            )

        # The old process kept the two started and processed activities
        self.assertEqual(old_process.activities.count(), 2)
        self.assertCountEqual(list(old_process.activities.values_list('state', flat=True).distinct()), [
            WorkerActivityState.Started,
            WorkerActivityState.Processed,
        ])

        # The new process gets the two queued and error activities in a started state as requested,
        # and also gets one extra activity for the act that had no activity before
        self.assertEqual(self.process.activities.count(), 3)
        self.assertCountEqual(list(self.process.activities.values_list('state', flat=True).distinct()), [
            WorkerActivityState.Started,
        ])

    def test_put_activity_requires_internal(self):
        """
        Only internal users (workers) are able to update the state of a worker activity
        Internal users with an instance admin are able to update a worker activity
        """
        internal_admin_user = User.objects.create_user('god@test.test', 'G0D')
        internal_admin_user.is_internal = True
        internal_admin_user.is_admin = True
        internal_admin_user.save()
        cases = (
            (None, status.HTTP_403_FORBIDDEN, 0),
            (self.user, status.HTTP_403_FORBIDDEN, 2),
            (self.superuser, status.HTTP_403_FORBIDDEN, 2),
            (self.internal_user, status.HTTP_200_OK, 5),
            (internal_admin_user, status.HTTP_200_OK, 5)
        )
        for user, status_code, requests_count in cases:
            self.activity.state = WorkerActivityState.Queued
            self.activity.save(update_fields=['state'])
            if user:
                self.client.force_login(user)
            with self.assertNumQueries(requests_count):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': WorkerActivityState.Started.value,
                    },
                    content_type='application/json',
                )
            self.assertEqual(response.status_code, status_code)

    def test_put_activity_wrong_worker_version(self):
        """
        Raises a generic error in case the worker version does not exists because a single SQL request is performed
        The response is a HTTP_409_CONFLICT
        """
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(uuid.uuid4())}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exist, is assigned to another process, or '
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_put_activity_element_unexisting(self):
        """
        Raises a generic error in case no activity exists for this element
        The response is a HTTP_409_CONFLICT
        """
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(uuid.uuid4()),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exist, is assigned to another process, or '
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_put_activity_process_unexisting(self):
        """
        Raises an error in case the process does not exist
        """
        self.client.force_login(self.internal_user)
        wrong_process_id = uuid.uuid4()
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(wrong_process_id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'process_id': [f'Invalid pk "{wrong_process_id}" - object does not exist.']
        })

    def test_put_activity_allowed_states(self):
        """
        Check each case state update is allowed depending on the actual state of the activity
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ['Queued', 'Started', 'Error', 'Processed']
        )
        allowed_states_update = (
            (queued, started, 5),
            (started, processed, 5),
            (started, error, 6),
            (error, queued, 5),
            (error, started, 5),
        )
        self.client.force_login(self.internal_user)
        for state, payload_state, nb_query in allowed_states_update:
            self.activity.state = state
            self.activity.save()
            with self.assertNumQueries(nb_query):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': payload_state,
                    },
                    content_type='application/json',
                )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.activity.refresh_from_db()
            self.assertEqual(self.activity.state.value, payload_state)

    def test_put_activity_forbidden_states(self):
        """
        Check state update is forbidden for some non consistent cases
        The response is a HTTP_409_CONFLICT
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ['Queued', 'Started', 'Error', 'Processed']
        )
        forbidden_states_update = (
            (queued, queued),
            (started, queued),
            (queued, error),
            (processed, error),
        )
        self.client.force_login(self.internal_user)
        for state, payload_state in forbidden_states_update:
            self.activity.state = state
            self.activity.save()
            with self.assertNumQueries(5):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': payload_state,
                    },
                    content_type='application/json',
                )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
            self.assertDictEqual(response.json(), {
                '__all__': [
                    'Either this worker activity does not exist, is assigned to another process, or '
                    f"updating the state to '{payload_state}' is forbidden."
                ]
            })
            self.activity.refresh_from_db()
            self.assertEqual(self.activity.state.value, state)

    # Just set the timeout to zero so that the transition always works
    @override_settings(WORKER_ACTIVITY_TIMEOUT=0)
    def test_put_activity_retry(self):
        self.client.force_login(self.internal_user)
        self.activity.state = WorkerActivityState.Started
        self.activity.save()

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    @override_settings(WORKER_ACTIVITY_TIMEOUT=10000)
    def test_put_activity_retry_limit(self):
        self.client.force_login(self.internal_user)
        self.activity.state = WorkerActivityState.Started
        self.activity.save()

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)

        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exist, is assigned to another process, or '
                "updating the state to 'started' is forbidden."
            ]
        })
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    def test_put_activity_process_overriding(self):
        """
        A process cannot update a worker activity even if it was initialized in another process
        """
        self.client.force_login(self.internal_user)
        process2 = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user
        )
        process2.worker_runs.create(version=self.worker_version, parents=[], configuration=self.configuration)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(process2.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exist, is assigned to another process, or '
                "updating the state to 'started' is forbidden."
            ]
        })
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.process_id, self.process.id)
        self.assertEqual(self.activity.state, WorkerActivityState.Queued)

    def test_put_activity_update_other_activities(self):
        """
        The error update should update the status on the following tasks
        """
        worker_version_2 = WorkerVersion.objects.get(worker__slug='dla')
        worker_version_3 = WorkerVersion.objects.get(worker__slug='worker-gpu')
        worker_version_4 = WorkerVersion.objects.create(
            worker=Repository.objects.first().workers.create(
                name='New version',
                slug='new',
                type=self.worker_type,
            ),
            revision=self.worker_version.revision,
            configuration={},
            state=WorkerVersionState.Available,
            docker_image=self.worker_version.docker_image
        )
        # Create runs run_1 > run_2 > run_3
        run_1 = self.process.worker_runs.first()
        run_2 = self.process.worker_runs.create(
            version=worker_version_2,
            parents=[run_1.id],
        )
        self.process.worker_runs.create(
            version=worker_version_3,
            parents=[run_2.id],
        )

        self.process.worker_runs.create(
            version=worker_version_4,
            parents=[],
        )

        # Create activities for run_2, run_3 and run_4
        activity_2 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_2,
            state=WorkerActivityState.Processed
        )
        activity_3 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_3,
            state=WorkerActivityState.Queued
        )
        activity_4 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_4,
            state=WorkerActivityState.Queued
        )

        self.client.force_login(self.internal_user)
        self.activity.state = WorkerActivityState.Started
        self.activity.save()

        with self.assertNumQueries(6):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Error.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Error)
        activity_2.refresh_from_db()
        self.assertEqual(activity_2.state, WorkerActivityState.Processed)
        activity_3.refresh_from_db()
        self.assertEqual(activity_3.state, WorkerActivityState.Error)
        activity_4.refresh_from_db()
        self.assertEqual(activity_4.state, WorkerActivityState.Error)

    @patch('arkindex.process.models.Process.versions')
    @patch('arkindex.process.models.ActivityManager.bulk_insert')
    def test_async_activities_error(self, bulk_insert_mock, versions_mock):
        """
        Handles errors while initializing workers activity of a process
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='volume'),
            activity_state=ActivityState.Pending
        )
        # Mock list_elements method to return the same queryset object
        elts_qs = Element.objects.filter(type__slug='volume')
        self.assertEqual(elts_qs.count(), 2)
        process.list_elements = MagicMock()
        process.list_elements.return_value = elts_qs

        worker_version_2 = WorkerVersion.objects.get(worker__slug='dla')
        process.worker_runs.create(version=self.worker_version, parents=[], configuration=None)
        process.worker_runs.create(version=worker_version_2, parents=[], configuration=None)

        bulk_insert_mock.side_effect = [
            bulk_insert_mock,
            Exception('Something went wrong')
        ]

        with self.assertRaises(Exception):
            initialize_activity(process)

        self.assertEqual(bulk_insert_mock.call_count, 2)
        self.assertCountEqual(bulk_insert_mock.call_args_list, [
            call(worker_version_id=self.worker_version.id, process_id=process.id, configuration_id=None, elements_qs=elts_qs),
            call(worker_version_id=worker_version_2.id, process_id=process.id, configuration_id=None, elements_qs=elts_qs)
        ])

        process.refresh_from_db()
        self.assertEqual(process.activity_state, ActivityState.Error)
        self.assertEqual(WorkerActivity.objects.filter(element__type__slug='volume').count(), 0)

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_corpus_acl(self):
        private_corpus = Corpus.objects.create(name="can't touch this")
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': private_corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have guest access to this corpus.'})

    def test_list_corpus_not_found(self):
        corpus_id = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
        self.assertNotEqual(str(self.corpus.id), corpus_id)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': corpus_id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'created': self.activity.created.isoformat().replace('+00:00', 'Z'),
                    'updated': self.activity.updated.isoformat().replace('+00:00', 'Z'),
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'worker_version_id': str(self.worker_version.id),
                    'configuration_id': str(self.configuration.id),
                    'state': 'queued'
                }
            ]
        })

    def test_list_filter_process_wrong_corpus(self):
        private_corpus = Corpus.objects.create(name="can't touch this")
        process = private_corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.superuser,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                {'process_id': str(process.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'process_id': ['This process does not exist.']})

    def test_list_filter_process(self):
        process = self.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.user,
        )
        activity = process.activities.create(
            element=self.element,
            worker_version=self.worker_version,
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(9):
            response = self.client.get(
                reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                {'process_id': str(process.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'created': activity.created.isoformat().replace('+00:00', 'Z'),
                    'updated': activity.updated.isoformat().replace('+00:00', 'Z'),
                    'element_id': str(self.element.id),
                    'process_id': str(process.id),
                    'worker_version_id': str(self.worker_version.id),
                    'configuration_id': None,
                    'state': 'queued'
                }
            ]
        })

    def test_list_filter_state(self):
        element = Element.objects.get(name='Volume 1')
        activity = element.activities.create(
            process=self.process,
            worker_version=self.worker_version,
            state=WorkerActivityState.Processed,
        )
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Processed).count(), 1)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                {'state': 'processed'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'created': activity.created.isoformat().replace('+00:00', 'Z'),
                    'updated': activity.updated.isoformat().replace('+00:00', 'Z'),
                    'element_id': str(element.id),
                    'process_id': str(self.process.id),
                    'worker_version_id': str(self.worker_version.id),
                    'configuration_id': None,
                    'state': 'processed'
                }
            ]
        })

    def test_list_invalid_filters(self):
        self.client.force_login(self.superuser)
        cases = [
            (
                {'process_id': 'a'},
                {'process_id': ['Process ID should be an UUID.']},
            ),
            (
                {'process_id': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'},
                {'process_id': ['This process does not exist.']},
            ),
            (
                {'state': 'lol'},
                {'state': ["'lol' is not a valid WorkerActivity state"]},
            ),
            (
                {'process_id': 'a', 'state': 'lol'},
                {
                    'process_id': ['Process ID should be an UUID.'],
                    'state': ["'lol' is not a valid WorkerActivity state"],
                },
            ),
        ]
        for filters, expected_errors in cases:
            with self.subTest(**filters):
                response = self.client.get(
                    reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                    filters
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)
