"""
Arkindex-specific Django system checks.

If you make any changes to those checks, you should consider updating the system
checks documentation on the wiki:
<https://wiki.vpn/en/arkindex/deploy/checks>
"""
import os
import sys

import yaml
from django.core.checks import Critical, Error, Warning, register

from arkindex.process.models import WorkerVersion
from ponos.recipe import parse_recipe


def only_runserver(func):
    "Decorator to run a system check only when running a dev server or deploying"
    from django.conf import settings

    def wrapper(*args, **kwargs):
        if settings.ALL_CHECKS:
            return func(*args, **kwargs)
        return []

    return wrapper


@register()
def api_urls_check(*args, **kwargs):
    '''
    Check that all urls end with a mandatory slash
    '''
    from arkindex.project.api_v1 import api
    return [
        Error(
            'This api url misses a mandatory ending /',
            hint=str(url),
            id='arkindex.E001',
        )
        for url in api
        if not str(url.pattern).endswith('/')
    ]


@register()
@only_runserver
def local_imageserver_check(*args, **kwargs):
    """
    Check that the configured local ImageServer exists
    """
    from django.conf import settings

    from arkindex.images.models import ImageServer
    local_id = settings.LOCAL_IMAGESERVER_ID

    try:
        ImageServer.objects.get(id=local_id)
    except ImageServer.DoesNotExist:
        return [Warning(
            f'Local ImageServer with ID {local_id} does not exist',
            hint=f'settings.LOCAL_IMAGESERVER_ID = {local_id}',
            id='arkindex.W009',
        )]

    return []


@register()
@only_runserver
def ponos_key_check(*args, **kwargs):
    """
    Warn about a missing Ponos private key that would prevent any Ponos agent from authenticating
    """
    from django.conf import settings
    if not os.path.exists(settings.PONOS_PRIVATE_KEY):
        return [Warning(
            f'Ponos private key at {settings.PONOS_PRIVATE_KEY} not found. '
            'Agents will be unable to connect to this server.',
            hint=f'`ponos.private_key` in {settings.CONFIG_PATH}',
            id='arkindex.W007',
        )]
    return []


@register()
def ponos_recipe_check(*args, **kwargs):
    """
    Check that the base Ponos recipe seems legit
    """
    from django.conf import settings
    errors = []

    recipe = settings.PONOS_RECIPE.copy()
    # Add a dummy task because Ponos wants at least one task
    recipe['tasks'] = {'task1': {'image': 'hello-world'}}
    try:
        parse_recipe(yaml.dump(recipe))
    except (ValueError, AssertionError) as e:
        errors.append(Error(
            'Base Ponos recipe validation failed: {}'.format(str(e)),
            hint='settings.PONOS_RECIPE',
            id='arkindex.E007',
        ))

    for variable in ('ARKINDEX_API_URL', 'ARKINDEX_API_TOKEN', 'ARKINDEX_API_CSRF_COOKIE'):
        if variable not in recipe.get('env', {}):
            errors.append(Warning(
                f'The {variable} environment variable should be defined '
                'to allow API client autoconfiguration in Ponos tasks',
                hint=f'`ponos.default_env` in {settings.CONFIG_PATH}',
                id='arkindex.W006',
            ))

    return errors


@register()
@only_runserver
def gitlab_oauth_check(*args, **kwargs):
    from django.conf import settings
    warnings = []
    app_id = getattr(settings, 'GITLAB_APP_ID', None)
    app_secret = getattr(settings, 'GITLAB_APP_SECRET', None)
    if not app_id:
        warnings.append(Warning(
            'GitLab app ID is not set; Git imports will fail.',
            hint='settings.GITLAB_APP_ID = {}'.format(repr(app_id)),
            id='arkindex.W003',
        ))
    if not app_secret:
        warnings.append(Warning(
            'GitLab app secret is not set; Git imports will fail.',
            hint='settings.GITLAB_APP_SECRET = {}'.format(repr(app_secret)),
            id='arkindex.W004',
        ))
    return warnings


@register()
@only_runserver
def s3_check(*args, **kwargs):
    from django.conf import settings
    aws_settings = {
        'AWS_ACCESS_KEY': 'AWS access key ID',
        'AWS_SECRET_KEY': 'AWS secret key',
        'AWS_THUMBNAIL_BUCKET': 'S3 thumbnails bucket name',
        'AWS_STAGING_BUCKET': 'S3 staging bucket name',
        'AWS_EXPORT_BUCKET': 'S3 export bucket name',
    }
    errors = []
    for name, display_name in aws_settings.items():
        value = getattr(settings, name, None)
        if not value:
            errors.append(Error(
                f'{display_name} is missing; all S3-related features will fail.',
                hint=f'settings.{name} = {value!r}',
                id='arkindex.E011',
            ))

    return errors


@register()
@only_runserver
def public_hostname_check(*args, **kwargs):
    from django.conf import settings
    if settings.PUBLIC_HOSTNAME is None:
        return [Warning(
            'The public_hostname setting is missing; absolute URLs may not be correctly generated.',
            hint='settings.PUBLIC_HOSTNAME',
            id='arkindex.W008',
        )]
    return []


@register()
def python_version_check(*args, **kwargs):
    if sys.version_info.minor < 8:
        return [
            Critical(
                'Arkindex requires Python 3.8 or later.',
                id='arkindex.C001',
            )
        ]
    return []


@register()
@only_runserver
def ingest_check(*args, **kwargs):
    from django.conf import settings
    warnings = []

    if settings.INGEST_S3_ACCESS_KEY is None:
        warnings.append(Warning(
            'INGEST_S3_ACCESS_KEY is not set. The S3 ingest feature will not work.',
            id='arkindex.W010',
        ))

    if settings.INGEST_S3_SECRET_KEY is None:
        warnings.append(Warning(
            'INGEST_S3_SECRET_KEY is not set. The S3 ingest feature will not work.',
            id='arkindex.W010',
        ))

    if settings.INGEST_S3_REGION is None and settings.INGEST_S3_ENDPOINT is None:
        warnings.append(Warning(
            'Neither INGEST_S3_REGION nor INGEST_S3_ENDPOINT are set. The S3 ingest feature will not work.',
            id='arkindex.W010',
        ))

    if settings.INGEST_IMAGESERVER_ID is None:
        warnings.append(Warning(
            'INGEST_IMAGESERVER_ID is not set. The S3 ingest feature will not work.',
            id='arkindex.W010',
        ))
    else:
        from arkindex.images.models import ImageServer
        try:
            ImageServer.objects.get(id=settings.INGEST_IMAGESERVER_ID)
        except ImageServer.DoesNotExist:
            warnings.append(Warning(
                f'ImageServer with ID {settings.INGEST_IMAGESERVER_ID} does not exist. The S3 ingest feature will not work.',
                hint=f'settings.INGEST_IMAGESERVER_ID = {settings.INGEST_IMAGESERVER_ID}',
                id='arkindex.W010',
            ))

    return warnings


@register()
def botocore_config_check(*args, **kwargs):
    """
    When not all S3 settings have both an access key ID, secret access key
    and either an endpoint or a region set, botocore might use some environment variables
    or configuration files to fill in the blanks. This can cause confusion for developers
    as some unit tests might fail or some other behaviors might be different, and can cause
    data leaks in production where some credentials might be used unexpectedly.
    """
    from django.conf import settings

    required_settings = [
        settings.AWS_ACCESS_KEY,
        settings.AWS_SECRET_KEY,
        settings.AWS_ENDPOINT or settings.AWS_REGION,
        settings.INGEST_S3_ACCESS_KEY,
        settings.INGEST_S3_SECRET_KEY,
        settings.INGEST_S3_ENDPOINT or settings.INGEST_S3_REGION,
    ]
    if all(setting is not None for setting in required_settings):
        # All settings are set, no complaints
        return []

    forbidden_variables = {
        'AWS_ACCESS_KEY_ID',
        'AWS_CONFIG_FILE',
        'AWS_CREDENTIAL_FILE',
        'AWS_DEFAULT_PROFILE',
        'AWS_DEFAULT_REGION',
        'AWS_EXECUTION_ENV',
        'AWS_PROFILE',
        'AWS_REGION',
        'AWS_SECRET_ACCESS_KEY',
        'AWS_SECURITY_TOKEN',
        'AWS_SESSION_TOKEN',
        'BOTO_CONFIG',
    }
    forbidden_files = {
        '~/.aws/config',
        '~/.aws/credentials',
    }
    warnings = []

    found_variables = forbidden_variables & set(os.environ.keys())
    warnings.extend([
        Warning(
            f'The {variable} environment variable is set, and not all S3 settings are set.\n'
            'This variable might override the Arkindex settings and cause unexpected behavior.',
            id='arkindex.W011',
        )
        for variable in found_variables
    ])

    for path in forbidden_files:
        if os.path.exists(os.path.expanduser(path)):
            warnings.extend([
                Warning(
                    f'The {path} file exists, and not all S3 settings are set.\n'
                    'The settings in this file might override the Arkindex settings and cause unexpected behavior.',
                    id='arkindex.W011',
                )
            ])

    return warnings


@register()
@only_runserver
def transkribus_check(*args, **kwargs):
    from django.conf import settings
    errors = []
    transkribus_feature = settings.ARKINDEX_FEATURES['transkribus']
    transkribus_config = {
        'TRANSKRIBUS_EMAIL': 'Transkribus account email address',
        'TRANSKRIBUS_PASSWORD': 'Transkribus account password',
        'TRANSKRIBUS_WORKER_VERSION': 'Transkribus worker version'
    }
    if not transkribus_feature:
        return []
    for setting, descr in transkribus_config.items():
        value = getattr(settings, setting)
        if not value:
            errors.append(Error(
                f'No {descr} is set; all Transkribus-related features will fail.',
                hint=f'settings.{setting} = {value!r}',
                id='arkindex.E012',
            ))
    worker_version = settings.TRANSKRIBUS_WORKER_VERSION
    if worker_version and not WorkerVersion.objects.filter(pk=worker_version).exists():
        errors.append(Error(
            'Transkribus worker version does not exist.',
            hint=f'settings.TRANSKRIBUS_WORKER_VERSION = {worker_version!r}',
            id='arkindex.E012',
        ))
    return errors
